/*
    File:   prj_05.cpp
    Author: Jonathan Dale Kirwan

    Creation Date:  Wed 03-Oct-2007 18:20:07
    Last Modified:  Wed 03-Oct-2007 18:20:07

    Copyright 2007, 2009, Jonathan Dale Kirwan, jonk@infinitefactors.org

    This source code module is distributed under the terms of the GNU General
    Public License (aka GNU GPL.)

    You should have received a copy of this license as COPYING.  But if not,
    it can currently be accessed over the web at:

        http://www.gnu.org/licenses/gpl.txt

    Or receive a copy by writing to the Free Software Foundation, Inc., the
    current address of which (at this time of writing) is:

        51 Franklin St
        Fifth Floor
        Boston, MA 02110
        USA


    DESCRIPTION

    This is project 5 -- set up a hardware timer in the MSP430 to replace the
    variable counter and make the LED blink on and off.

    In the previous project, a "busy wait" loop used a counter variable to set
    the delay time.  Larger numbers resulted in longer delays and that led to
    slower LED blinking rates.  A busy wait loop is very easy to understand
    and very quick to implement, so it is often used when there is a "quick
    and dirty" need for a delay.  It is _not_ often used in the final version
    of commercial products, however.  (Except in rare cases I won't go into
    right now.)  A hardware resource called a 'timer' is often used, instead.

    One problem with using a software-only "busy wait" loop is that the delay
    it creates depends upon a simple count-down or count-up method.  Different
    microcontrollers, or even the same ones used differently, do not all run
    at the same instruction execution rates.  Just like IBM PCs, which have
    800MHz versions and 3GHz versions and will run games better or worse,
    depending; the same idea also applies to microcontrollers.  Some of them
    run faster than others.  So if you use a counter to set some delay, it's
    almost certain that it won't run the same way on every microcontroller,
    even if it is the exact same model.  Electronics designers are free to set
    the speed as they like.  (Software designers can also sometimes affect the
    speed, too, in microcontrollers like the MSP430.)  So software programmers
    need to be aware of this and attempt to write code that provides consistent
    results, if and when that's possible.

    Luckily, most microcontrollers include a special-function unit called a
    "timer."  These aren't much different that what we've been doing in
    software with a counter variable and a busy loop:  a special counter is
    added into the hardware and it counts down (or up) from some value and
    when the value gets to zero, the timer takes another step to let you know
    about that fact.  Having the hardware do this for you, though, allows your
    software to do other things in the meantime.  (Most microcontroller
    manufacturers have added a variety of other features to their timers.  They
    usually do more than just counting up and down.  So they are often very,
    very useful and sometimes quite fancy devices.  But just about all of them
    can count up and down.  It's a good idea to read the datasheet to see what
    other abilities they have and think about them for your software use.)

    Using a hardware timer, if available in the microcontroller, requires some
    extra code designed to properly "configure it."  This is because a timer
    usually supports more than one operational mode and you need to tell it
    which one of several modes to use.  Also, as with the counter variable used
    in the busy wait loop, the hardware counter needs to be initialized with a
    starting value to specify the delay.  The upshot here is that some careful
    reading of the datasheet or its product family manual is required to know
    how to do it, properly.

    It's for this reason that the earlier busy wait loop done entirely in soft-
    ware is easier to implement.  Just write code in the c/c++ language and
    compile it.  But two downsides remain.

    First, the actual delay strongly depends upon the code generated by the
    c/c++ compiler for the loop.  Different compilers, and even different
    compile options for the same compiler, will produce different delays and
    thus require changing the initial counter value to "fine tune" them to the
    desired time.  Using a hardware timer resource can help solve this problem
    because the rate at which the counter is updated (counted up or counted
    down) is fixed by the hardware itself and how you configured it and doesn't
    depend on the code that the compiler generates.

    The actual delay also depends on the processor speed, too.  Both software-
    only and hardware timer methods suffer here.  But hardware timers usually
    include some help by including "divisors" which can be used to tweak the
    actual count rate in the hardware so that similar counter values can still
    be used.  So the hardware provides some simple aids, though it's not a
    perfect solution because you still need to make those adjustments just as
    you would with a software-only busy loop.

    Second, and at least equally important if not more so, hardware timers may
    completely free up the microcontroller's cpu to perform other useful work
    while the hardware timer performs its job, in parallel.  To fully achieve
    this goal, though, hardware interrupts are usually used together with the
    hardware timer resource.  And this project will NOT do that, quite yet.

    We could make a big leap from Project 4 and solve both complaints of the
    software busy wait loop in this project.  But I'm going to introduce the
    hardware timer counter, first, and use it to replace some of the use of a
    software counter variable and just solve the first complaint mentioned
    above -- the dependence of the delay upon compiler and compiler options.
    Projects 6 and 7 will then solve the second complaint, in two more steps.
    At the end of Project 7, the hardware timer will be used in a way that
    frees up the microcontroller for other useful work while the LED blinks at
    some desired rate.

    Luckily, most microcontrollers include a special-function unit called a
    "timer."  These aren't much different that what we've been doing in
    software with a counter variable and a busy loop:  a special counter is
    added into the hardware and it counts down (or up) from some value and
    when the value gets to zero, the timer takes another step to let you know
    about that fact.  Having the hardware do this for you, though, allows your
    software to do other things in the meantime.  (Most microcontroller
    manufacturers have added a variety of other features to timers.  They
    usually do more than just counting up and down.  So they are often very,
    very useful and sometimes quite fancy devices.  But just about all of them
    can count up and down.  It's a good idea to read the datasheet to see what
    other abilities they have and think about them for your software use.)

    In this example, despite the fact that the timer can be arranged to do even
    more helpful things, we will keep it simple to start out.  The main thing
    here is to add code to set up the timer and to monitor it's counter.

    Read the comments near the code below and modify the program so that the
    LED blinks at an even rate you can see, but using the timer this time.

    Once you've finished making a version that blinks the LED on and off, try
    setting up different counts in the timer and see what happens.

    So, here are the steps for project 5:

        (1) Compile the initial version included here and verify that
            the LED blinks at 1Hz.  I've used 50 times 10 milliseconds
            between LED changes, which should work out to one blink
            per second.  Check this out and verify it.

        (2) Modify the loop variable used to count the number of 10
            milliceond periods.  It's currently at 50.  Try smaller
            and larger values.  Does this do what you expect?

        (3) Put the loop variable back at 50, again, after step (2),
            and now find the value of 19999 and modify that to a
            different number.  Such as 10000 or 30000.  Run that and
            see how it behaves.  Can you explain that behavior, as
            well?


    TIMER A2 DETAILS

    It's probably best to include a diagram of how Timer A2 is being used.
    The uppper diagram helps show the timer counter's upward motion, over
    time.  The line made with text, showing peaks and valleys, represents
    the timer counter, itself, as it counts upwards.  The timer automatically
    starts counting upwards when the counter reaches zero and automatically
    resets back to zero when the counter reaches a value that is set into
    TACCR0.  In this application, TACCR0 is set just once and isn't adjusted
    afterwards.  It sets the basic cycle time for the LED pulsing.

            TACCR0 ----> ..^......^......^......^......^.....
                          /|     /|     /|     /|     /|     /   Upward
                         / |    / |    / |    / |    / |    /    Counter
                        /  |   /  |   /  |   /  |   /  |   /     Diagram
                    |  /   |  /   |  /   |  /   |  /   |  / 
                    | /    | /    | /    | /    | /    | /   
                    |/     |/     |/     |/     |/     |/     
            ZERO ...v......v......v......v......v......v.....

                ----|------|------|------|------|------|---
                    T      T      T      T      T      T   
                    A      A      A      A      A      A      Event Names
                    I      I      I      I      I      I      as Described
                    F      F      F      F      F      F      in the User's
                    G      G      G      G      G      G      Guide from TI

            time ----> ----> ----> ----> ----> ----> ----> ---->


    NOTE

    In these projects, I will sometimes place long discussions in the code.
    These discussions will be highlighted by "*** READ THIS ***".  It's wise
    to pay particular attention to these sections and read the text there.  In
    later projects, I'll remove comments that duplicate what was already said
    in earlier ones.  So check back on earlier projects, if needed.


    TARGET COMPILER

    This module is designed to be compiled with the IAR C++/C compiler that
    comes on a CD with the Texas Instruments "MSP320 Ultra-Low-Power MCUs
    eZ430-F2013 Development Tool."  This IAR compiler is also known as the
    "IAR Kickstart" compiler and it can be downloaded from the web site for
    Texas Instruments, on pages concerning the MSP430 processor.

    IAR provides the Kickstart version of their compiler, but imposes some
    limitations on your code size to encourage you to buy their full compiler
    license if you try and use it for professional work.  But for the purposes
    of education here, the IAR Kickstart tools for the MSP430 are sufficient.
 
 
    MODIFICATIONS
 
    Original source.
*/


/*  ----------------------------------------------------------------------  */
#include <msp430x20x3.h>
/*  ----------------------------------------------------------------------
    Include definitions appropriate for the MSP430F2013 microcontroller.
    It is part of the compiler vendor's product, designed to describe key
    datasheet details, and was not written separately for this project.

    See the file itself for details, under the installation directory for
    the IAR compiler toolset.
*/


/*  ----------------------------------------------------------------------  */
int main( ) {
/*  ----------------------------------------------------------------------
    Project 5 starts here.  See discussion in the file header for detailed
    information about the project.
*/

    /*  Turn the watchdog timer off.  See Project 1 for more info.
    */
        WDTCTL= WDTPW | WDTHOLD;

    /*  Enable port pin P1.0 for output (P1.0 is the LED pin.)  See Project
        1 for more info.
    */
        P1DIR |= 0x01;

    /*  Set up the processor to run at a calibrated 16MHz rate.  This will
        probably vary a little over temperature, but it should be close.
        See Project 3 for more info.
    */
        BCSCTL1= CALBC1_16MHZ;
        DCOCTL= CALDCO_16MHZ;

    /*  *** READ THIS ***
        The MSP430F2013 comes standard with a timer called "timer_A2."  The
        way Texas Instruments does this is to name their timer types by a 
        letter.  So this is timer type A.  Whatever that means.  They also
        have a timer type B, but that one isn't included with the MSP430F2013.
        Anyway, the number afterwards is the number of comparison values you
        can use.  Since this is an A2 timer, it's a type A that includes two
        comparison values.  (These two comparison values can also be used
        instead to take a snapshot of the timer as it is counting away.  But
        that's for another time.)

        Here, and you'd need to read the documentation from TI to gather the
        details well on this, I'm setting the timer up for our use.  It's
        being "configured" for a pre-divide-by-8, or a 2MHz count rate if the
        microcontroller was properly set to its calibrated 16MHz, above.  The
        counter is 16 bits in size, so that means the counter will roll-over
        fairly often (longest time possible is 65536*0.5 microseconds, or no
        longer than 32.768 milliseconds.  That would be a blinking rate of
        some 30 times per second -- still a bit fast.  So I plan to set this
        up for one timer event per 10 milliseconds (that is well within its
        range) and then count up the number of 10 millisecond events to set
        the blinking rate.  (The value 19999 is actually 1-less than the
        actual count that takes place, so it really does count 20000 cycles.)
    */
        TACTL= MC_0 | TASSEL_2 | ID_3 | TACLR;      // reset it, first.
        TACCTL0= CM_0 | CCIS_2 | SCS | OUTMOD_1;    // set the mode.
        TACCR0= 19999;                              // 10ms intervals.
        TACTL= MC_1 | TASSEL_2 | ID_3;              // turn it on.

    /*  *** READ THIS ***
        Now we are ready to blink the LED.  Since we are going to keep on
        doing this forever, we can use that same "forever" for loop we've
        already been using.  However, we are now going to count 50 events
        so that each change in the LED will take place on 500 millisecond
        intervals.  (Remember, the timer flag is being set every 10 milli-
        seconds.)  And each time the timer flag is set to 1, we need to
        clear it.

        This should blink the LED at a rate of 2*500 milliseconds or once per
        second.  Try calling up the Windows clock and checking!
    */

        for ( ; ; ) {
            P1OUT ^= 0x01;                      // using XOR toggles LED
            for ( int i= 0; i < 50; ++i ) {
                while ( (TACTL & TAIFG) == 0 )  // wait for the flag to set
                    ;
                TACTL &= ~TAIFG;                // clear the flag now
            }
        }

    return 0;
}
