/*
    File:   prj_06.cpp
    Author: Jonathan Dale Kirwan

    Creation Date:  Wed 03-Oct-2007 18:20:07
    Last Modified:  Wed 03-Oct-2007 18:20:07

    Copyright 2007, 2009, Jonathan Dale Kirwan, jonk@infinitefactors.org

    This source code module is distributed under the terms of the GNU General
    Public License (aka GNU GPL.)

    You should have received a copy of this license as COPYING.  But if not,
    it can currently be accessed over the web at:

        http://www.gnu.org/licenses/gpl.txt

    Or receive a copy by writing to the Free Software Foundation, Inc., the
    current address of which (at this time of writing) is:

        51 Franklin St
        Fifth Floor
        Boston, MA 02110
        USA


    DESCRIPTION

    This is project 6 -- setting up a timer, with interrupt events, to help
    make the LED blink on and off.

    In the previous project, you used a "busy wait" loop to set the delay time
    between LED on/off changes.  This works, but it usually isn't good practice
    because it ties up the microcontroller doing just that one thing.  Most
    real-world projects need to do several things at once and cannot afford to
    set aside everything else while merely blinking an LED.

    In addition, there is another problem with using a "busy wait" loop that
    depends on a simple count-down or count-up method.  Microcontrollers do not
    all run at the same speed.  Just like IBM PCs, which have 800MHz versions
    and 3GHz versions and will run games better or worse, depending; the same
    thing applies to microcontrollers, as well.  Some of them run faster than
    others.  So if you use a counter to set some delay, it's almost certain
    that it won't run the same way on every microcontroller, even if it is the
    exact same model.  Electronics designers are free to set the speed as they
    like.  So software programmers need to be aware of this and attempt to
    write compensating code, if possible.

    Luckily, most microcontrollers include a special-function unit called a
    "timer."  These aren't much different that what we've been doing in
    software with a counter variable and a busy loop:  a special counter is
    added into the hardware and it counts down (or up) from some value and
    when the value gets to zero, the timer takes another step to let you know
    about that fact.  Having the hardware do this for you, though, allows your
    software to do other things in the meantime.  (Most microcontroller
    manufacturers have added a variety of other features to timers.  They
    usually do more than just counting up and down.  So they are often very,
    very useful and sometimes quite fancy devices.  But just about all of them
    can count up and down.  It's a good idea to read the datasheet to see what
    other abilities they have and think about them for your software use.)

    In this example, despite the fact that the timer can be arranged to do even
    more helpful things, we will keep it simple to start out.  The main thing
    here is to add code to set up the timer and to monitor it's counter.

    Read the comments near the code below and modify the program so that the
    LED blinks at an even rate you can see, but using the timer this time.

    Once you've finished making a version that blinks the LED on and off, try
    setting up different counts in the timer and see what happens.

    So, here are the steps for project 6:

        (1) Compile the initial version included here and verify that
            the LED blinks at 1Hz.  I've used 50 times 10 milliseconds
            between LED changes, which should work out to one blink
            per second.  Check this out and verify it.

        (2) Modify the loop variable used to count the number of 10
            milliceond periods.  It's currently at 50.  Try smaller
            and larger values.  Essentially, just duplicate what was
            done in Project 5.  The main thing is to make sure this
            does exactly the same thing using a different approach.

    This is a shorter tutorial, mainly to get started using interrupts.  If
    you get through this quickly, Project 7 will put on the finishing touches.


    NOTE

    In these projects, I will sometimes place long discussions in the code.
    These discussions will be highlighted by "*** READ THIS ***".  It's wise
    to pay particular attention to these sections and read the text there.  In
    later projects, I'll remove comments that duplicate what was already said
    in earlier ones.  So check back on earlier projects, if needed.


    TARGET COMPILER

    This module is designed to be compiled with the IAR C++/C compiler that
    comes on a CD with the Texas Instruments "MSP320 Ultra-Low-Power MCUs
    eZ430-F2013 Development Tool."  This IAR compiler is also known as the
    "IAR Kickstart" compiler and it can be downloaded from the web site for
    Texas Instruments, on pages concerning the MSP430 processor.

    IAR provides the Kickstart version of their compiler, but imposes some
    limitations on your code size to encourage you to buy their full compiler
    license if you try and use it for professional work.  But for the purposes
    of education here, the IAR Kickstart tools for the MSP430 are sufficient.
 
 
    MODIFICATIONS
 
    Original source.
*/


/*  ----------------------------------------------------------------------  */
#include <msp430x20x3.h>
/*  ----------------------------------------------------------------------
    Include definitions appropriate for the MSP430F2013 microcontroller.
    It is part of the compiler vendor's product, designed to describe key
    datasheet details, and was not written separately for this project.

    See the file itself for details, under the installation directory for
    the IAR compiler toolset.
*/


static volatile bool timer_event;   // see comments below

/*  ----------------------------------------------------------------------  */
int main( ) {
/*  ----------------------------------------------------------------------
    Project 6 starts here.  See discussion in the file header for detailed
    information about the project.
*/

    /*  Turn the watchdog timer off.  See Project 1 for more info.
    */
        WDTCTL= WDTPW | WDTHOLD;

    /*  Enable port pin P1.0 for output (P1.0 is the LED pin.)  See Project
        1 for more info.
    */
        P1DIR |= 0x01;

    /*  Set up the processor to run at a calibrated 16MHz rate.  This will
        probably vary a little over temperature, but it should be close.
        See Project 3 for more info.
    */
        BCSCTL1= CALBC1_16MHZ;
        DCOCTL= CALDCO_16MHZ;

    /*  *** READ THIS ***
        Set up timer_a for 10 millisecond intervals and enable interrupt
        events.  There is a slight difference between this code and the
        code used in Project 5.  Do you see it?  (The CCIE bit, perhaps?)

        The CCIE bit in the TACCTL0 register instructs the timer to try
        and generate an interrupt event each time the CCIFG bit gets set.
        The CCIFG bit will be automatically cleared when the interrupt
        event is actually serviced (by entering the designated function),
        so explicit software code isn't needed to clear it so long as we
        permit the interrupting event.  This permission not only requires
        the CCIE bit but also requires the general interrupt system to be
        enabled (disabling it means the entire interrupt system is turned
        off.)  Other than this CCIE bit, the code here is the same as in
        Project 5.
    */
        TACTL= MC_0 | TASSEL_2 | ID_3 | TACLR;          // reset it, first.
        TACCTL0= CM_0 | CCIS_2 | SCS | OUTMOD_1 | CCIE; // set the mode.
        TACCR0= 19999;                                  // 10ms intervals.
        TACTL= MC_1 | TASSEL_2 | ID_3;                  // turn it on.

    /*  *** READ THIS ***
        In the past few projects, we didn't need to support interrupts taking
        place.  All control was inside the main() function and we didn't allow
        the microcontroller to jump around when events took place.  In fact,
        we didn't allow any events, before.  Now, we need to allow the timer
        hardware to force a subroutine to be executed every so often.  This
        happens, hidden behind the scenes.  But it happens.  However, we need
        to explicitly tell the microcontroller that it is okay to allow these
        sudden interrupts to take place.  This library function does that.  I
        could have done this differently, but the special function here is
        the recommended method.  Sometimes, there are changes in the hardware
        from one supported microcontroller to another and using the library
        call allows the compiler to determine the right method given the
        target microcontroller.

        Once interrupts are enabled, the timer can force our code to start
        executing a different part of the code when the timer interval has
        completed.  You will see the special interrupt subroutine below main().

        (This library function yields just two machine instructions.)
    */
        __enable_interrupt( );

    /*  *** READ THIS ***
        Now we are ready to blink the LED.  Just as it was in Project 5, we
        are going to keep on doing this forever and use the same "forever"
        for-loop we've already been using.  And just as before, we are still
        using a counter here to extend the timer_A2 device's period by
        counting 50 events so that each change in the LED will take place on
        500 millisecond intervals.  (The timer is interrupting every 10ms.)

        The difference here is that instead of directly observing and then
        clearing a particular timer flag within the code here, we will use
        a for-loop that waits for the interrupt routine to 'signal' that an
        interrupt took place.  Basically, this is the same method except we
        are using an interrupt routine and the built-in interrupt mechanism
        instead of directly observing the flag, waiting for it, and then
        clearing it, manually.

        Note that the timer_event variable is qualified by the 'volatile'
        keyword (look earlier in the code to see it.)  This is because the
        variable can be arbitrarily modified in the timer_A2 interrupt
        event code (last function in this file) and the for-loop here
        needs to know that fact.  However, the 'i' variable doesn't need
        to be qualified as volatile.

        This should blink the LED at a rate of 2*500 milliseconds or once per
        second.  Try calling up the Windows clock and checking!
    */
        for ( ; ; ) {
            P1OUT ^= 0x01;              // using XOR toggles the pin value.
            for ( int i= 0; i < 50; ++i )
                for ( timer_event= false; !timer_event; )
                    ;
        }

    return 0;
}


/*  ----------------------------------------------------------------------  */
#pragma vector= TIMERA0_VECTOR
__interrupt void timer_A2( void ) {
/*  ----------------------------------------------------------------------
    This routine is called whenever the timer reaches the special count
    that was set up, earlier.  When we get here, we just set the special
    timer_event variable to signal that the event happened and then quit.
    That way, the code in main() can take it from there.

    *** READ THIS***

    The above #pragma is a special compiler command.  Actually, there are
    lots of commands that can follow a #pragma and different c/c++ com-
    pilers will treat them differently.  So this is a special area where
    you need to read the documentation on the compiler, itself, to under-
    stand these in thorough detail.  This one tells the c++ compiler that
    the function that follows it is supposed to be called whenever there
    is a timer event.  This is handled via something called a "vector
    table" on the MSP430F2013, and this explains why theey use a keyword,
    "vector," above.  However, just accept that this works.  It's a
    necessary extra.
*/
        timer_event= true;

    return;
}
