/*
    File:   prj_07.cpp
    Author: Jonathan Dale Kirwan

    Creation Date:  Wed 03-Oct-2007 18:20:07
    Last Modified:  Wed 03-Oct-2007 18:20:07

    Copyright 2007, 2009, Jonathan Dale Kirwan, jonk@infinitefactors.org

    This source code module is distributed under the terms of the GNU General
    Public License (aka GNU GPL.)

    You should have received a copy of this license as COPYING.  But if not,
    it can currently be accessed over the web at:

        http://www.gnu.org/licenses/gpl.txt

    Or receive a copy by writing to the Free Software Foundation, Inc., the
    current address of which (at this time of writing) is:

        51 Franklin St
        Fifth Floor
        Boston, MA 02110
        USA


    DESCRIPTION

    This is project 7 -- extend the timer interrupt code to completely off-
    load the work in making the LED blink on and off.  Also, learn about one
    of the low power modes.

    In the previous project, you still used a "busy wait" loop that depended
    upon the value of a variable set by an interrupt routine triggered by the
    timer_A2 device.  This works, but as mentioned before it still isn't good
    practice.  Most real-world projects need to do several things at once and
    cannot afford to set aside everything else while merely blinking an LED.

    We made some advancements in the right direction in the earlier two
    projects.  One was to start using the hardware timer_A2 device and the
    other was to start learning to use interrupt events in writing our code.
    This time we take one more important logical step and completely free
    the main code from having to do anything, at all, leaving all of the work
    for the interrupt routine, itself.  This completely frees up the main
    code.  Of course, in this tutorial we will still just waste that benefit
    by doing nothing at all in the main code.  But we could do something else
    such as computing PI for fun or doing some useful calculation.

    So, here are the steps for project 7:

        (1) Compile the initial version included here and verify that
            the LED blinks at 1Hz.  I've used 50 times 10 milliseconds
            between LED changes, which should work out to one blink
            per second.  Check this out and verify it.

        (2) Examine the differences in the code between Project 6 and
            this one.  Do they make sense to you?  Notice that the
            earlier volatile file-scope variable has been replaced by
            a local function scope, less qualified static variable in
            the interrupt code, itself.

        (3) Try changing the initial value of 50 in the first line of
            code in the interrupt function to 1 or 100.  Can you tell
            any difference?  Now try 0.  Can you explain this well?

        (4) Set the initial value back to 50, but now go into the if-
            block of code in the interrupt function and change the
            value of 50 there to something larger.  Does the change in
            behavior make sense to you?  Try commenting that line out
            and see what happens.  Make sense?

        (5) Note that in compiling this code (and in most of the steps
            of earlier projects), there is a single warning message
            about unreachable code.  Find the two #pragma statements
            that are commented out near the end of the main() function
            and remove the leading // characters and place the # as the
            first character of the line for each.  Now recompile.  You
            should see that there are no warnings, anymore.  Do you
            recall why the warning was there in the first place?  And
            why we can accept it and order the compiler to stop giving
            that message?

        (6) Comment out the two lines marked "PART A" and uncomment
            the line marked "PART B" in the code below.  Compile and
            test.  Works the same?  (Well, it should... but using a
            lot less power, now.)  Now go ahead and entirely remove
            the #pragma's involved in step (5).  Note that the warning
            doesn't re-appear.  Why?

    After the last step above, we have made a lot of progress.  The micro-
    controller is now operating in a very low power mode while simultaneously
    blinking the LED, and doing all that in the background so that the main
    code could be doing something entirely different and useful.  This is
    quite a lot and a big step forward in developing useful applications.


    NOTE

    In these projects, I will sometimes place long discussions in the code.
    These discussions will be highlighted by "*** READ THIS ***".  It's wise
    to pay particular attention to these sections and read the text there.  In
    later projects, I'll remove comments that duplicate what was already said
    in earlier ones.  So check back on earlier projects, if needed.


    TARGET COMPILER

    This module is designed to be compiled with the IAR C++/C compiler that
    comes on a CD with the Texas Instruments "MSP320 Ultra-Low-Power MCUs
    eZ430-F2013 Development Tool."  This IAR compiler is also known as the
    "IAR Kickstart" compiler and it can be downloaded from the web site for
    Texas Instruments, on pages concerning the MSP430 processor.

    IAR provides the Kickstart version of their compiler, but imposes some
    limitations on your code size to encourage you to buy their full compiler
    license if you try and use it for professional work.  But for the purposes
    of education here, the IAR Kickstart tools for the MSP430 are sufficient.
 
 
    MODIFICATIONS
 
    Original source.
*/


/*  ----------------------------------------------------------------------  */
#include <msp430x20x3.h>
/*  ----------------------------------------------------------------------
    Include definitions appropriate for the MSP430F2013 microcontroller.
    It is part of the compiler vendor's product, designed to describe key
    datasheet details, and was not written separately for this project.

    See the file itself for details, under the installation directory for
    the IAR compiler toolset.
*/


/*  ----------------------------------------------------------------------  */
int main( ) {
/*  ----------------------------------------------------------------------
    Project 7 starts here.  See discussion in the file header for detailed
    information about the project.
*/

    /*  Turn the watchdog timer off.  See Project 1 for more info.
    */
        WDTCTL= WDTPW | WDTHOLD;

    /*  Enable port pin P1.0 for output (P1.0 is the LED pin.)  See Project
        1 for more info.
    */
        P1DIR |= 0x01;

    /*  Set up the processor to run at a calibrated 16MHz rate.  This will
        probably vary a little over temperature, but it should be close.
        See Project 3 for more info.
    */
        BCSCTL1= CALBC1_16MHZ;
        DCOCTL= CALDCO_16MHZ;

    /*  Set up timer_a for 10 millisecond intervals and enable interrupts.
    */
        TACTL= MC_0 | TASSEL_2 | ID_3 | TACLR;          // reset it, first.
        TACCTL0= CM_0 | CCIS_2 | SCS | OUTMOD_1 | CCIE; // set the mode.
        TACCR0= 19999;                                  // 10ms intervals.
        TACTL= MC_1 | TASSEL_2 | ID_3;                  // turn it on.

    /*  Enable interrupt events, globally.  (Which, of course, also allows
        the ones we care about -- timer_A2's.)
    */
        __enable_interrupt( );

    /*  *** READ THIS ***
        Now we are ready to blink the LED.  Just as it was in Project 6, we
        are going to keep on doing this forever and use the same "forever"
        for-loop we've already been using.  And just as before, we are still
        using a counter here to extend the timer_A2 device's period by
        counting 50 events so that each change in the LED will take place on
        500 millisecond intervals.  (The timer is interrupting every 10ms.)

        The difference here is that all of the monitoring code that used to
        be in the main routine has been moved into the interrupt function.

        This should blink the LED at a rate of 2*500 milliseconds or once per
        second.  Try calling up the Windows clock and checking!
    */
        for ( ; ; )                     // PART A
            ;                           // PART A

    /*  *** READ THIS ***
        The following line switches the processor over into a deep, low power
        mode.  This can be used instead of the above loop marked PART A.  It
        effectively stops the processor from executing until an interrupt
        event takes place.  When the interrupt takes place, the fact that the
        microcontroller was earlier stopped is saved away and then restored
        when the interrupt finishes.  So the microcontroller merely returns
        to its stopped mode after the interrupt code is done.  Thus, it never
        continues on to the 'return' statement below it -- the same as in all
        the other, earlier projects.
    */
        // __low_power_mode_0( );          // PART B

// #pragma diag_suppress=Pe111
    return 0;
// #pragma diag_warning=Pe111
}


/*  ----------------------------------------------------------------------  */
#pragma vector= TIMERA0_VECTOR
__interrupt void timer_A2( void ) {
/*  ----------------------------------------------------------------------
    This routine is called whenever the timer reaches the special count
    that was set up, earlier.  The routine accepts 50 interrupts before
    changing the LED state and then resets the counter so that it will
    take another 50 for the next change.  All of the work required to
    properly blink the LED is located here, freeing the main code to do
    other things.
*/
    static int counter= 50;

        if ( --counter == 0 ) {
            P1OUT ^= 0x01;              // using XOR toggles the pin value
            counter= 50;                // reset the counter for next time
        }

    return;
}
