/*
    File:   prj_08.cpp
    Author: Jonathan Dale Kirwan

    Creation Date:  Wed 03-Oct-2007 18:20:07
    Last Modified:  Wed 03-Oct-2007 18:20:07

    Copyright 2007, 2009, Jonathan Dale Kirwan, jonk@infinitefactors.org

    This source code module is distributed under the terms of the GNU General
    Public License (aka GNU GPL.)

    You should have received a copy of this license as COPYING.  But if not,
    it can currently be accessed over the web at:

        http://www.gnu.org/licenses/gpl.txt

    Or receive a copy by writing to the Free Software Foundation, Inc., the
    current address of which (at this time of writing) is:

        51 Franklin St
        Fifth Floor
        Boston, MA 02110
        USA


    DESCRIPTION

    This is project 8 -- learning about duty cycles other than 50% by using
    the up-down timer_A2 mode.

    After Project 7, the LED is being handled entirely by the use of a timer
    and an interrupt routine, freeing the main code for other work.  The way
    the code is designed in Project 7 (and earlier projects, as well), the
    off and on LED time periods are equal.  This amounts to a 50% duty cycle,
    which simply means that the on-time is one half of the sum of the on and
    off times for the LED.

    In this project, it's time to make adjustments to this duty cycle and see
    how to program that and what the results look like, too.

    Here are the steps for project 8:

        (1) Compile the initial version included here and verify that
            the LED blinks at 1Hz.  The interrupt code has been
            changed to use an up/down counter mode and a comparator
            value that determines when the LED changes state.  Check
            this out and make sure that it mimics Project 7's behavior.
            Examine the differences in the code between Project 7 and
            this one.  Do they make sense to you?

        (2) Note the initial values of 'counter' and 'count_direction'.
            Is it possible that counter may be set, at first, to a
            negative value if DUTY is 0?  Should it still work okay?
            Change DUTY to 0 and observe.  Change DUTY to the same
            value that PERIOD has and also observe.  Does the behavior
            make sense?  Now move DUTY to various other values from 0
            to the value in PERIOD (it is fine to use 0 and PERIOD or
            any value in between them.)  Does what happens make sense?

        (3) Try setting PERIOD to 4 and DUTY to 2.  You should still be
            able to see the blinking effect.  But it's faster.  Also,
            since DUTY can only be from 0 to 4, the duty cycle of the
            LED is limited to 0%, 25%, 50%, 75%, and 100%.  The limited
            resolution is the price paid by using a long timer interval
            that is 10 milliseconds long.  We could move PERIOD to 8 and
            DUTY could be any of nine values, then, giving twice the
            duty cycle resolution.  But then the blinking rate would be
            slower, too.

        (4) Find the TACCR0 setting, where it uses 19999, and change the
            value to 9999, instead.  Now set PERIOD to 8 and DUTY to 4.
            Compile and run.  You should see the same blinking rate that
            appeared in step (3) when PERIOD was 4 and DUTY was 2.  But
            now there is more duty cycle resolution.  Do you understand
            what happened?  (I changed the basic timer event from 10ms
            to 5ms!  Then a period of 8 is the same time as the older
            period of 4 was.)  The price paid by having a faster timer
            event period that provides better duty cycle resolution is
            the sheer number of times that the interrupt code executes.
            Each execution takes up cpu time.  And this takes away from
            cpu time available elsewhere.  No free lunches!


    NOTE

    In these projects, I will sometimes place long discussions in the code.
    These discussions will be highlighted by "*** READ THIS ***".  It's wise
    to pay particular attention to these sections and read the text there.  In
    later projects, I'll remove comments that duplicate what was already said
    in earlier ones.  So check back on earlier projects, if needed.


    TARGET COMPILER

    This module is designed to be compiled with the IAR C++/C compiler that
    comes on a CD with the Texas Instruments "MSP320 Ultra-Low-Power MCUs
    eZ430-F2013 Development Tool."  This IAR compiler is also known as the
    "IAR Kickstart" compiler and it can be downloaded from the web site for
    Texas Instruments, on pages concerning the MSP430 processor.

    IAR provides the Kickstart version of their compiler, but imposes some
    limitations on your code size to encourage you to buy their full compiler
    license if you try and use it for professional work.  But for the purposes
    of education here, the IAR Kickstart tools for the MSP430 are sufficient.
 
 
    MODIFICATIONS
 
    Original source.
*/


/*  ----------------------------------------------------------------------  */
#include <msp430x20x3.h>
/*  ----------------------------------------------------------------------
    Include definitions appropriate for the MSP430F2013 microcontroller.
    It is part of the compiler vendor's product, designed to describe key
    datasheet details, and was not written separately for this project.

    See the file itself for details, under the installation directory for
    the IAR compiler toolset.
*/


/*  ----------------------------------------------------------------------  */
int main( ) {
/*  ----------------------------------------------------------------------
    Project 8 starts here.  See discussion in the file header for detailed
    information about the project.
*/

    /*  Turn the watchdog timer off.  See Project 1 for more info.
    */
        WDTCTL= WDTPW | WDTHOLD;

    /*  Enable port pin P1.0 for output (P1.0 is the LED pin.)  See Project
        1 for more info.
    */
        P1DIR |= 0x01;

    /*  Set up the processor to run at a calibrated 16MHz rate.  This will
        probably vary a little over temperature, but it should be close.
        See Project 3 for more info.
    */
        BCSCTL1= CALBC1_16MHZ;
        DCOCTL= CALDCO_16MHZ;

    /*  Set up timer_a for 10 millisecond intervals and enable interrupts.
    */
        TACTL= MC_0 | TASSEL_2 | ID_3 | TACLR;          // reset it, first.
        TACCTL0= CM_0 | CCIS_2 | SCS | OUTMOD_1 | CCIE; // set the mode.
        TACCR0= 19999;                                  // 10ms intervals.
        TACTL= MC_1 | TASSEL_2 | ID_3;                  // turn it on.

    /*  Enable interrupt events, globally.  (Which, of course, also allows
        the ones we care about -- timer_A2's.)
    */
        __enable_interrupt( );

    /*  Enter sleep mode and allow the interrupt code to do all the work.
    */
        __low_power_mode_0( );

    return 0;
}


#define PERIOD      (50)
#define DUTY        (25)

/*  ----------------------------------------------------------------------  */
#pragma vector= TIMERA0_VECTOR
__interrupt void timer_A2( void ) {
/*  ----------------------------------------------------------------------
    This routine performs an up/down counter function with a comparator
    that may match the counter value and, if it does match, drives the LED
    either on or off based upon the direction of the up/down counter at
    the time.
*/
    static int counter= DUTY - 1;
    static enum { up, down } count_direction= up;

    /*  Adjust the counter up or down, as indicated, and change direction
        when either 0 or the PERIOD is reached so that the counter is always
        at or between these values.
    */
        switch ( count_direction ) {
        case up:
            if ( ++counter >= PERIOD )
                count_direction= down;
            break;
        case down:
            if ( --counter <= 0 )
                count_direction= up;
            break;
        }

    /*  If there is a counter match with the value set by DUTY, the LED should
        either be turned on or off, depending on the current count direction.
    */
        if ( counter == DUTY )
            switch ( count_direction ) {
            case up:
                P1OUT |= 0x01;          // LED on
                break;
            case down:
                P1OUT &= ~0x01;         // LED off
                break;
            }

    return;
}
