/*
    File:   prj_09.cpp
    Author: Jonathan Dale Kirwan

    Creation Date:  Wed 03-Oct-2007 18:20:07
    Last Modified:  Wed 03-Oct-2007 18:20:07

    Copyright 2007, 2009, Jonathan Dale Kirwan, jonk@infinitefactors.org

    This source code module is distributed under the terms of the GNU General
    Public License (aka GNU GPL.)

    You should have received a copy of this license as COPYING.  But if not,
    it can currently be accessed over the web at:

        http://www.gnu.org/licenses/gpl.txt

    Or receive a copy by writing to the Free Software Foundation, Inc., the
    current address of which (at this time of writing) is:

        51 Franklin St
        Fifth Floor
        Boston, MA 02110
        USA


    DESCRIPTION

    This is project 9 -- using varying duty cycles with fast periods to
    simulate brightness changes in LEDs.

    In Project 8, we added code in the timer event to provide an up/down
    counter and a comparator value we could use to set the duty cycle of
    the LED pulsing.  The period of the LED blinking was set by PERIOD
    and the duty cycle was set by DUTY, in Project 8.  But timer_A2 was
    operated with a basic time period of 10 milliseconds.

    In this project, timer_A2 will be sped up a lot so that we can experiment
    with rates above the human eye's critical flicker fusion frequency (CFF)
    at 100% modulation.  I mentioned the 100% modulation because that's all
    we can do with the LED -- turn it on or off.  There are different rates
    at which the human eye will "fuse" flickering light sources.  The rate
    partly depends on the depth of modulation and partly depends on the light
    intensity level.  But we don't care much about how those factors affect
    things because we can't control them.  We can only use 100% modulation
    and the intensity is already defined, as well, by the eZ430-F2013 kit.
    So granting those facts and assuming the worst (very intense, if you are
    looking closely at the USB stick), the CFF is pretty close to 70-80 Hz.
    This means we need to operate the LED there or faster.

    As already noted (hopefully) many projects back, the LED we have on this
    eZ430-F2013 kit is attached to pin P1.0 and it's digital.  So we can turn
    the LED full-on or else turn it off.  There is no in between.  (This, the
    "100% modulation" already mentioned.)  So although the intent in Project 9
    is to manually vary the apparent brightness, there is no apparent means to
    do that except for this concept of the CFF in human visual perception.
    Luckily (or unluckily), it turns out that human vision "integrates" light
    that flickers faster than the CFF.  So if the LED is pulsed very fast, it
    will "look steady."  You can still see the pulses by using an oscilloscope
    and you can know they are there, but you won't see it no matter how hard
    you look if it is blinking fast enough.

    By now, it should be crystal clear where I'm headed.  We can use this fact
    of the CFF to advantage.  If we blink the LED on and off a bit faster, say
    70Hz more or less, most folks won't see the LED blinking but will instead
    see it somewhat dimmer if the duty cycle is less than 100%.  If we blink
    it at 100Hz, this means 10 milliseconds per complete cycle that includes
    both the ON and OFF times.  If we turn the LED on for 5ms and turn it off
    for the other 5ms, then we have achieved a 100Hz blinking rate.  But it is
    is only emitting light for half the time.  So our eyes pick this up as
    "somewhat dimmer."  If we turned it on for 1ms and turned it off for 9ms,
    then it would appear still dimmer.  We can hold the flicker rate fixed
    (100Hz, for example), and just vary the time we keep the LED on within
    each cycle.  The fact that humans won't see the blinking and will instead
    turn that into a brightness sensation can be used in our circumstance.

    And that's what we will do.

    Doing this is called "varying the duty cycle" or "pulse width modulation"
    or PWM.  It is a term you will learn to recognize, if you haven't already.
    If you have a cycle time of 10ms and you turn on the LED for 4ms and off
    for 6ms, then your 'on' duty cycle is said to be 40%.  And if you keep the
    period (sum of off and on times) fixed, while adjusting the duty cycle,
    then it is pulse width modulation at some given frequency (the inverse of
    the period.)

    There are many uses for PWM.  Don't assume it only applies to LEDs.

    In this project, we make adjustments to interrupt rate and duty cycle
    to experiment with how the LED appears when it's period is much shorter.

    Here are the steps for project 9:

        (1) Compile the initial version included here and verify that
            the LED blinks at about the same starting rate as it did
            in Project 8.

        (2) Modify the line where TACCR0 is set to 19999.  Now make it
            199, instead.  This will set the time to 100 microseconds,
            instead of 10 milliseconds.  Keep PERIOD at (50) and set
            DUTY to (25) and recompile and test the program.  You should
            see the LED appear 'steady' on.  However, it isn't.  It is
            blinking.  Just at a much faster rate, now.  We still have
            a 50% duty cycle, though, because DUTY is half of PERIOD.

        (3) Modify DUTY to try out different values.  For example, set
            it to 5, 45, and 48 and test each of these.  Do you notice
            that when DUTY is 48, that the LED appears significantly
            dimmer?  Try 49, now.  Note that we aren't actually changing
            the brightness of the LED when it is on.  We can't, in fact.
            All we can do is turn it on or off.  But our eye averages
            (same meaning as 'integrates' here) what it sees when
            flickering is very fast like this.  So we perceive it as
            dimmer.

            (A CRT television, in fact, depends on this aspect of your
            vision.  If you had faster visual response times, you'd be
            able to actually see a bright dot traversing the screen,
            and you wouldn't see much of a picture, at all.)

        (4) If you can, find a USB cable that allows you to place the
            USB stick in your hand where you can wave it around.  Set
            DUTY to 0, which should turn the LED without flickering.
            Compile and run it.  Turn out the lights or get into a dark
            environment and wave your hand back and forth at a fast, but
            constant rate.  You should see a continuous green arc appear
            where you are waving it back and forth.  Now change DUTY to
            48, recompile and run.  Again, turn out the lights and wave
            your hand back and forth.  You should see the individual
            locations where the LED is on, separated by substantial parts
            of the arc that remain dark to the eye.  You can now see the
            fact that the LED is blinking, by spreading it out in space.
            From this you should retain at least three facts:  (a) a rate
            of about 100Hz or faster makes the LED appear steady; and (b)
            changing the duty cycle lets you adjust the apparent bright-
            ness; and (c) LEDs that are in motion may require a still
            higher frequency to appear steady.  Don't assume 100Hz is
            enough for a steady appearance if the LED is on vibrating
            machinery!!


    NOTE

    In these projects, I will sometimes place long discussions in the code.
    These discussions will be highlighted by "*** READ THIS ***".  It's wise
    to pay particular attention to these sections and read the text there.  In
    later projects, I'll remove comments that duplicate what was already said
    in earlier ones.  So check back on earlier projects, if needed.


    TARGET COMPILER

    This module is designed to be compiled with the IAR C++/C compiler that
    comes on a CD with the Texas Instruments "MSP320 Ultra-Low-Power MCUs
    eZ430-F2013 Development Tool."  This IAR compiler is also known as the
    "IAR Kickstart" compiler and it can be downloaded from the web site for
    Texas Instruments, on pages concerning the MSP430 processor.

    IAR provides the Kickstart version of their compiler, but imposes some
    limitations on your code size to encourage you to buy their full compiler
    license if you try and use it for professional work.  But for the purposes
    of education here, the IAR Kickstart tools for the MSP430 are sufficient.
 
 
    MODIFICATIONS
 
    Original source.
*/


/*  ----------------------------------------------------------------------  */
#include <msp430x20x3.h>
/*  ----------------------------------------------------------------------
    Include definitions appropriate for the MSP430F2013 microcontroller.
    It is part of the compiler vendor's product, designed to describe key
    datasheet details, and was not written separately for this project.

    See the file itself for details, under the installation directory for
    the IAR compiler toolset.
*/


/*  ----------------------------------------------------------------------  */
int main( ) {
/*  ----------------------------------------------------------------------
    Project 9 starts here.  See discussion in the file header for detailed
    information about the project.
*/

    /*  Turn the watchdog timer off.  See Project 1 for more info.
    */
        WDTCTL= WDTPW | WDTHOLD;

    /*  Enable port pin P1.0 for output (P1.0 is the LED pin.)  See Project
        1 for more info.
    */
        P1DIR |= 0x01;

    /*  Set up the processor to run at a calibrated 16MHz rate.  This will
        probably vary a little over temperature, but it should be close.
        See Project 3 for more info.
    */
        BCSCTL1= CALBC1_16MHZ;
        DCOCTL= CALDCO_16MHZ;

    /*  Set up timer_a for 10 millisecond intervals and enable interrupts.
        We will change this interval rate in step (2).
    */
        TACTL= MC_0 | TASSEL_2 | ID_3 | TACLR;          // reset it, first.
        TACCTL0= CM_0 | CCIS_2 | SCS | OUTMOD_1 | CCIE; // set the mode.
        TACCR0= 19999;   /* make 199 in step (2) */     // 10ms intervals.
        TACTL= MC_1 | TASSEL_2 | ID_3;                  // turn it on.

    /*  Enable interrupt events, globally.  (Which, of course, also allows
        the ones we care about -- timer_A2's.)
    */
        __enable_interrupt( );

    /*  Enter sleep mode and allow the interrupt code to do all the work.
    */
        __low_power_mode_0( );

    return 0;
}


#define PERIOD      (50)
#define DUTY        (25)

/*  ----------------------------------------------------------------------  */
#pragma vector= TIMERA0_VECTOR
__interrupt void timer_A2( void ) {
/*  ----------------------------------------------------------------------
    This routine performs an up/down counter function with a comparator
    that may match the counter value and, if it does match, drives the LED
    either on or off based upon the direction of the up/down counter at
    the time.
*/
    static int counter= DUTY - 1;
    static enum { up, down } count_direction= up;

    /*  Adjust the counter up or down, as indicated, and change direction
        when either 0 or the PERIOD is reached so that the counter is always
        at or between these values.
    */
        switch ( count_direction ) {
        case up:
            if ( ++counter >= PERIOD )
                count_direction= down;
            break;
        case down:
            if ( --counter <= 0 )
                count_direction= up;
            break;
        }

    /*  If there is a counter match with the value set by DUTY, the LED should
        either be turned on or off, depending on the current count direction.
    */
        if ( counter == DUTY )
            switch ( count_direction ) {
            case up:
                P1OUT |= 0x01;          // LED on
                break;
            case down:
                P1OUT &= ~0x01;         // LED off
                break;
            }

    return;
}
