/*
    File:   prj_10.cpp
    Author: Jonathan Dale Kirwan

    Creation Date:  Wed 03-Oct-2007 18:20:07
    Last Modified:  Wed 03-Oct-2007 18:20:07

    Copyright 2007, 2009, Jonathan Dale Kirwan, jonk@infinitefactors.org

    This source code module is distributed under the terms of the GNU General
    Public License (aka GNU GPL.)

    You should have received a copy of this license as COPYING.  But if not,
    it can currently be accessed over the web at:

        http://www.gnu.org/licenses/gpl.txt

    Or receive a copy by writing to the Free Software Foundation, Inc., the
    current address of which (at this time of writing) is:

        51 Franklin St
        Fifth Floor
        Boston, MA 02110
        USA


    DESCRIPTION

    This is project 10 -- replacing explicit up/down counter code in the
    interrupt handling function with the timer_A2 up/down count mode.

    In Project 8, we added code in the timer event to provide an up/down
    counter and a comparator value we could use to set the duty cycle of
    the LED pulsing.  The period of the LED blinking was set by PERIOD
    and the duty cycle was set by DUTY, in Project 8.  But timer_A2 was
    operating in an upward count mode.

    In Project 9, we experimented with the human eye's critical flicker
    fusion frequency (CFF), setting the flicker rate above the rate at
    which we can see the blinking.

    In this project, we use an up/down count mode in timer_A2 rather than
    the upward counting mode, maintaining the high flicker rate of the LED.
    This way, we can remove some code in the timer interrupt.  The timer
    includes a special compare register we can use, TACCR1, in very much
    the same way that we earlier used DUTY in Project 9.  The difference
    is that the timer has a special register where we can set that value.

    Note the change in a #pragma used just before the interrupt function.
    Compare it to a similar one found in Project 9.  The reason for the
    difference has to do with how the MSP430 was designed -- a compare match
    with TACCR1 causes a different interrupt vector to be used than in our
    earlier cases where the interrupt event depended upon a match with TACCR0.

    Also, this project introduces a greater use of pre-processor commands.
    You may modify LEDRATE and SYSCLK.  Note that there are some computations
    involved.  The main reason for setting things up like this is to bring to
    the top of the source code a couple of interesting parameters you can
    change and to then cause the code to adapt to those requirements.  See if
    you can follow the calculations themselves and what they do.

    Finally, note that the interrupt routine is much shorter than before.
    In this case, all that happens is to toggle the LED each time the routine
    is entered for a CCIFG1 event.  The timer handles the rest of the work.

    Here are the steps for project 10:

        (1) Compile the initial version included here and verify that
            the LED appears solid on.  It should be running at 50% duty
            cycle, but at a fast 100Hz blinking rate.  (Note the 0.50
            multiplier in the code.)

        (2) Modify the value of DUTY from 0.50 to 0.997 (about 0.3% ON
            time.)  Compile and test.  Does the LED appear dimmer?  Try
            other values for DUTY and see if they make some sense.  Note
            that it's not linear in the sense that 0.25 is not twice as
            bright as 0.50... in fact, it might be barely noticed.  More
            on that -- read about the Talbot-Plateau Law, later on here.

        (3) Try some different values and also try that experiment from
            Project 9 where the USB stick is waved around in a darkened
            room.  See if the behavior is sufficiently similar (in other
            words, proves that this new timer method works about as well
            as the earlier software-only method.)

        (4) Try keeping SYSCLK at 16000000 but changing LEDRATE to 10.
            Compile and an error should result.  Change LEDRATE to 15.
            Still an error.  Try 16.  Should work.  This is because the
            clock rate of the CPU is so high that the timer_A2 16-bit
            register just isn't big enough for very slow rates.  So now
            try setting SYSCLK to 1000000 and change LEDRATE to 4.  It
            should compile fine and run approximately as expected.  Note
            that if you still wanted an LEDRATE of 4, but also wanted to
            keep SYSCLK at 16000000, then you'd need to add software to
            help extend things.  (Like Project 9.)


    TIMER A2 DETAILS

    It's probably best, once again, to include a diagram of how Timer A2 will
    now be used.  The upper part of the diagram helps show the timer counter's
    up-down motion over time.  The line made with text, showing peaks and
    valleys, represents the timer counter, itself, as it counts upwards and
    then downwards.  This up/down motion of the counter will be allowed to
    run freely, without interrupting our code.  It will simply operate in the
    background, so to speak.  However, we will set up a comparator value in
    TACCR1 and we _do_ want to get interrupts based upon when that comparison
    generates a match.

            TACCR0 ---->   ^             ^             ^
                          / \           / \           / \        Up-Down
                         /   \         /   \         /   \       Counter
            TACCR1 ...../.....\......./.....\......./.....\...   Diagram
                 \     /       \     /       \     /       \
                  \   /         \   /         \   /         \
                   \ /           \ /           \ /           \
            ZERO    v             v             v

                 -------|-----|-------|-----|-------|-----|---
                        C     C       C     C       C     C
                        C     C       C     C       C     C      Event Names
                        I     I       I     I       I     I      as Described
                        F     F       F     F       F     F      in the User's
                        G     G       G     G       G     G      Guide from TI
                        1     1       1     1       1     1
                 -------|-----|-------|-----|-------|-----|---

            ON          ,-----,       ,-----,       ,-----,
                        |     |       |     |       |     |      LED
                        |     |       |     |       |     |      State
            OFF    -----'     '-------'     '-------'     '---

            time ----> ----> ----> ----> ----> ----> ----> ---->


    TALBOT-PLATEAU LAW

    This law in the field of visual perception states that if a light source
    blinks fast enough that it passes beyond the critical flicker fusion
    frequency (CFF), then the brightness will appear to be the same "as if"
    the light source were steadily operated at the time-averaged luminance.
    In other words, if you operate the light source at twice the luminance
    but only half the time (50% on and 50% off, flickering faster than the
    CFF), then it will appear to have the same luminance as that similar
    light source operated at the lower luminance.

    Keep in mind that operating a light at a 50% duty cycle, above the CFF,
    does not mean that it will appear half as bright as one operated 100%
    of the time.  It means that operating a light at 50% duty cycle and faster
    than the CFF will make it appear the same brightness as another operating
    at half the luminence.  There is a difference in meaning.  Graple with it.

    The reason it doesn't look "half as bright" when operating at a 50% duty
    cycle is that the human eye works according to another law, the Weber-
    Fechner law.  That one states that intensity goes as the logarithm of the
    physical stimuli.

    Well, almost.  There's a pair of other ones, the Broca-Sulzer effect and
    the Bruecke-Bartley effect, that operate at flicker frequencies below the
    CFF.  At around 50ms on-duration and 50ms off-duration, for example, a
    blinking light source will appear to be brighter than the time-averaged
    luminance ... in apparent violation of the Talbot-Plateau law.  But the
    Talbot-Plateau law doesn't apply at such low frequencies, so it isn't
    really a violation.

    Actually, there are a lot of "laws" in the field of visual optics and
    optometry.  These are just three or so out of many more.  But important
    ones, perhaps, to know about.  And like many fields where the science is
    still being hashed out (the brain does a lot of processing we don't yet
    fully understand), laws are assumed to be more approximation than physical
    fact.  So don't be surprised to find these laws broken from time to time.


    NOTE

    In these projects, I will sometimes place long discussions in the code.
    These discussions will be highlighted by "*** READ THIS ***".  It's wise
    to pay particular attention to these sections and read the text there.  In
    later projects, I'll remove comments that duplicate what was already said
    in earlier ones.  So check back on earlier projects, if needed.


    TARGET COMPILER

    This module is designed to be compiled with the IAR C++/C compiler that
    comes on a CD with the Texas Instruments "MSP320 Ultra-Low-Power MCUs
    eZ430-F2013 Development Tool."  This IAR compiler is also known as the
    "IAR Kickstart" compiler and it can be downloaded from the web site for
    Texas Instruments, on pages concerning the MSP430 processor.

    IAR provides the Kickstart version of their compiler, but imposes some
    limitations on your code size to encourage you to buy their full compiler
    license if you try and use it for professional work.  But for the purposes
    of education here, the IAR Kickstart tools for the MSP430 are sufficient.
 
 
    MODIFICATIONS
 
    Original source.
*/


/*  ----------------------------------------------------------------------  */
#include <msp430x20x3.h>
/*  ----------------------------------------------------------------------
    Include definitions appropriate for the MSP430F2013 microcontroller.
    It is part of the compiler vendor's product, designed to describe key
    datasheet details, and was not written separately for this project.

    See the file itself for details, under the installation directory for
    the IAR compiler toolset.
*/


/*  ----------------------------------------------------------------------  */
#define DUTY        (0.50)
/*  ----------------------------------------------------------------------
    This value selects the LED duty cycle.  0.5 is equivalent to 50%.  The
    value specifies what percent of the time that the LED is on.  So using
    0.1 will cause the LED to be dimmer than using 0.8.
*/


/*  ----------------------------------------------------------------------  */
#define LEDRATE     (100)           // in Hertz
/*  ----------------------------------------------------------------------
    Choose an LED pulse rate here.  I recommend something at 70Hz or
    faster, since the LED shouldn't appear to the eye as blinking.  But
    other rates are just fine.  Tinker.
*/


/*  ----------------------------------------------------------------------  */
#define SYSCLK      (16000000)      // in Hertz
/*  ----------------------------------------------------------------------
    Defines the basic SYSCLK rate we want to use here.  Use calibrated
    values only, unless you are willing to set up the DCO and BC1 values
    and create the appropriate SYSCLK value for them.  For now, an error
    is generated if SYSCLK isn't one of the standard calibration values.

    If you want to add an uncalibrated mode here, you need to provide some
    definitions in the #else clause where the #error statement is found.
*/
#if SYSCLK == 16000000
# define DCO         (CALDCO_16MHZ)
# define BC1         (CALBC1_16MHZ)
#elif SYSCLK == 12000000
# define DCO         (CALDCO_12MHZ)
# define BC1         (CALBC1_12MHZ)
#elif SYSCLK == 8000000
# define DCO         (CALDCO_8MHZ)
# define BC1         (CALBC1_8MHZ)
#elif SYSCLK == 1000000
# define DCO         (CALDCO_1MHZ)
# define BC1         (CALBC1_1MHZ)
#else
# error Uncalibrated SYSCLK requested.
#endif


/*  ----------------------------------------------------------------------
    COMPUTE USEFUL PREPROCESSOR SYMBOLS
    ----------------------------------------------------------------------
    Don't change the following preprocessor statements unless you know
    what you are doing and understand their current design.  They are
    here to compute proper programming values for registers used in
    configuring the timer_A2 device.  An error is generated if the
    LEDRATE and SYSCLK cannot be simultaneously accomodated.

    The purpose of each is as follows:
        TICKS       number of SYSCLK ticks per LED pulse period
        IDVAL       the SYSCLK pre-divisor to use; must be power of 2
        IDCTL       the TACCTL field value needed to achieve IDVAL
        TAC         1 plus the required TACCR0 value
*/
#define TICKS       (SYSCLK/2/LEDRATE)
#define IDVAL       (1+(TICKS >> 16))
#if IDVAL > 4
# undef IDVAL
# define IDVAL      (8)
# define IDCTL      (ID_3)
#elif IDVAL > 2
# undef IDVAL
# define IDVAL      (4)
# define IDCTL      (ID_2)
#elif IDVAL == 2
# define IDCTL      (ID_1)
#elif IDVAL == 1
# define IDCTL      (ID_0)
#else
# error LEDRATE is negative; try a positive value.
#endif
#if (TICKS/IDVAL) > 65535
# error LEDRATE is too small for the given SYSCLK; try a lower SYSCLK.
#endif
#define TAC         ((unsigned int) (TICKS/IDVAL))
#define TACLINE     ((unsigned int) (TICKS*DUTY/IDVAL))


/*  ----------------------------------------------------------------------  */
int main( ) {
/*  ----------------------------------------------------------------------
    Project 10 starts here.  See discussion in the file header for
    detailed information about the project.
*/

    /*  Turn the watchdog timer off.  See Project 1 for more info.
    */
        WDTCTL= WDTPW | WDTHOLD;

    /*  Enable port pin P1.0 for output (P1.0 is the LED pin.)  See Project
        1 for more info.
    */
        P1DIR |= 0x01;

    /*  *** READ THIS ***
        Set up the processor to run at a calibrated rate.  These values are
        computed earlier (see above.)  They are chosen to reflect the SYSCLK
        value.
    */
        BCSCTL1= BC1;
        DCOCTL= DCO;

    /*  *** READ THIS ***
        Set up timer_a for appropriate intervals and then enable interrupt
        events.  This time, we are setting up timer_A2 for up/down count
        mode and using a special comparator value found in TACCR1.  Read
        the documentation from Texas Instruments on the timer, too!
    */
        TACTL= MC_0 | TASSEL_2 | IDCTL | TACLR;         // reset it, first.
        TACCTL0= CM_0 | CCIS_2 | SCS | OUTMOD_1;        // set the
        TACCTL1= CM_0 | CCIS_2 | SCS | OUTMOD_1 | CCIE; //    modes.
        TACCR0= TAC - 1u;                               // upward limit
        TACCR1= TACLINE;                                // sets duty cycle
        TACTL= MC_3 | TASSEL_2 | IDCTL;                 // turn it on.

    /*  Enable interrupt events, globally, and go to sleep.
    */
        __enable_interrupt( );
        __low_power_mode_0( );

    return 0;
}


#define     CCIFG1_EVENT    (0x02)

/*  ----------------------------------------------------------------------  */
#pragma vector= TIMERA1_VECTOR
__interrupt void timer_A2( void ) {
/*  ----------------------------------------------------------------------
    This routine reads the TAIV register to find out what caused the
    interrupt.  If it is a CCIFG1 compare-match event, then the LED is
    toggled to create the flicker effect.
*/
        if ( TAIV == CCIFG1_EVENT )
            P1OUT ^= 0x01;

    return;
}
